/*
jGuard is a security framework based on top of jaas (java authentication and authorization security).
it is written for web applications, to resolve simply, access control problems.
version $Name$
http://sourceforge.net/projects/jguard/

Copyright (C) 2004  Charles GAY

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


jGuard project home page:
http://sourceforge.net/projects/jguard/

*/
package net.sf.jguard.ext.java5.authentication.jmx;

import java.io.ObjectInputStream;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.security.AccessControlException;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.logging.Logger;

import javax.management.Attribute;
import javax.management.AttributeList;
import javax.management.AttributeNotFoundException;
import javax.management.InstanceAlreadyExistsException;
import javax.management.InstanceNotFoundException;
import javax.management.IntrospectionException;
import javax.management.InvalidAttributeValueException;
import javax.management.ListenerNotFoundException;
import javax.management.MBeanException;
import javax.management.MBeanInfo;
import javax.management.MBeanPermission;
import javax.management.MBeanRegistration;
import javax.management.MBeanRegistrationException;
import javax.management.MBeanServer;
import javax.management.MBeanTrustPermission;
import javax.management.MalformedObjectNameException;
import javax.management.NotCompliantMBeanException;
import javax.management.NotificationFilter;
import javax.management.NotificationListener;
import javax.management.ObjectInstance;
import javax.management.ObjectName;
import javax.management.OperationsException;
import javax.management.QueryExp;
import javax.management.ReflectionException;
import javax.management.loading.ClassLoaderRepository;
import javax.management.remote.MBeanServerForwarder;

import net.sf.jguard.core.authorization.policy.LocalAccessController;

/**
 * 
 * @author <a href="mailto:diabolo512@users.sourceforge.net">Charles Gay</a>
 *
 */
public class MBeanServerGuard implements MBeanServerForwarder {
	private static final Logger logger = Logger.getLogger(MBeanServerGuard.class.getName());
	private MBeanServer mbs = null;
	private LocalAccessController accessController = null;
	
	public MBeanServerGuard(LocalAccessController lac){
		accessController = lac;
	}
	
	public MBeanServer getMBeanServer() {
		return mbs;
	}

	public void setMBeanServer(MBeanServer mBeanServer) {
		mbs = mBeanServer;
	}

	public void addNotificationListener(ObjectName name,
			NotificationListener listener, NotificationFilter filter,
			Object handback) throws InstanceNotFoundException {
		accessController.checkPermission(new MBeanPermission(getClassName(name),null,name,"addNotificationListener"));
		mbs.addNotificationListener(name, listener, filter, handback);

	}

	public void addNotificationListener(ObjectName name, ObjectName listener,
			NotificationFilter filter, Object handback)
			throws InstanceNotFoundException {
		accessController.checkPermission(new MBeanPermission(getClassName(name),null,name,"addNotificationListener"));
		mbs.addNotificationListener(name, listener, filter, handback);
	}

	public ObjectInstance createMBean(String className, ObjectName name)
			throws ReflectionException, InstanceAlreadyExistsException,
			MBeanRegistrationException, MBeanException,
			NotCompliantMBeanException {
		
		return  createMBean(className, name, (Object[]) null, (String[]) null);
		
	}

	public ObjectInstance createMBean(String className, ObjectName name,
			ObjectName loaderName) throws ReflectionException,
			InstanceAlreadyExistsException, MBeanRegistrationException,
			MBeanException, NotCompliantMBeanException,
			InstanceNotFoundException {
		return createMBean(className, name, loaderName, (Object[]) null, (String[]) null);
	}

	public ObjectInstance createMBean(String className, ObjectName name,
			Object[] params, String[] signature) throws ReflectionException,
			InstanceAlreadyExistsException, MBeanRegistrationException,
			MBeanException, NotCompliantMBeanException {
		ObjectInstance oi = null;
		
		try {
			return createMBean(className, name, (ObjectName)null, (Object[]) null, (String[]) null);
		} catch (InstanceNotFoundException e) {
			logger.severe(e.getMessage());
		}
		return oi;
		
	}

	public ObjectInstance createMBean(String className, ObjectName name,
			ObjectName loaderName, Object[] params, String[] signature)
			throws ReflectionException, InstanceAlreadyExistsException,
			MBeanRegistrationException, MBeanException,
			NotCompliantMBeanException, InstanceNotFoundException {
		//instantiate permission check
		accessController.checkPermission(new MBeanPermission(className,null,null,"instantiate"));
		
		//register permission check
		accessController.checkPermission(new MBeanPermission(className,null,name,"registerMBean"));
		Class clazz = null;
		try {
			clazz = Thread.currentThread().getContextClassLoader().loadClass(className);
		} catch (ClassNotFoundException e) {
			logger.severe(e.getMessage());
		}
		 if(!clazz.getProtectionDomain().implies(new MBeanTrustPermission("register"))){
			 throw new AccessControlException("registration denied");
		 }
		
		if(name==null){
			Class[] classes = new Class[signature.length];
			for (int i =0;i<signature.length;i++) {
				String element = signature[i];
				try {
					classes[i] = Thread.currentThread().getContextClassLoader().loadClass(element);
				} catch (ClassNotFoundException e) {
					logger.severe(e.getMessage());
				}
			}
			Constructor constructor;
			Object obj = null;
			try {
				constructor = clazz.getDeclaredConstructor(classes);
				obj = constructor.newInstance(params);
			} catch (SecurityException e) {
				logger.severe(e.getMessage());
			} catch (NoSuchMethodException e) {
				logger.severe(e.getMessage());
			} catch (IllegalArgumentException e) {
				logger.severe(e.getMessage());
			} catch (InstantiationException e) {
				logger.severe(e.getMessage());
			} catch (IllegalAccessException e) {
				logger.severe(e.getMessage());
			} catch (InvocationTargetException e) {
				logger.severe(e.getMessage());
			}
			
			MBeanRegistration mbeanReg = (MBeanRegistration)obj;
			try {
				name = (ObjectName)mbeanReg.preRegister(mbs, null);
			} catch (Exception e) {
				logger.severe(e.getMessage());
			}
			accessController.checkPermission(new MBeanPermission(getClassName(name),null,name,"registerMBean"));
		}
		
		 
		ObjectInstance oi = mbs.createMBean(className, name);
		return oi;
	}

	public ObjectInputStream deserialize(ObjectName name, byte[] data)
			throws InstanceNotFoundException, OperationsException {
		accessController.checkPermission(new MBeanPermission(getClassName(name),null,name,"getClassLoaderFor"));
		return mbs.deserialize(name, data);
	}

	public ObjectInputStream deserialize(String className, byte[] data)
			throws OperationsException, ReflectionException {
		accessController.checkPermission(new MBeanPermission(null,null,null,"getClassLoaderRepository"));
		return mbs.deserialize(className, data);
	}

	public ObjectInputStream deserialize(String className,
			ObjectName loaderName, byte[] data)
			throws InstanceNotFoundException, OperationsException,
			ReflectionException {
		accessController.checkPermission(new MBeanPermission(getClassName(loaderName),null,loaderName,"getClassLoader"));
		return mbs.deserialize(className, loaderName, data);
	}

	public Object getAttribute(ObjectName name, String attribute)
			throws MBeanException, AttributeNotFoundException,
			InstanceNotFoundException, ReflectionException {
		accessController.checkPermission(new MBeanPermission(getClassName(name),attribute,name,"getAttribute"));
		return mbs.getAttribute(name, attribute);
	}

	public AttributeList getAttributes(ObjectName name, String[] attributes)
			throws InstanceNotFoundException, ReflectionException {
		accessController.checkPermission(new MBeanPermission(getClassName(name),null,name,"getAttribute"));
		AttributeList attList = mbs.getAttributes(name, attributes);
		for(int i = 0;i<attList.size();i++){
			Attribute att = (Attribute)attList.get(i);
			try{
				accessController.checkPermission(new MBeanPermission(getClassName(name), att.getName(), name, "getAttribute"));
			}catch(AccessControlException ace){
				attList.remove(att);
				i--;
			}
		}
		return attList;
	}

	public ClassLoader getClassLoader(ObjectName loaderName)
			throws InstanceNotFoundException {
			accessController.checkPermission(new MBeanPermission(getClassName(loaderName),null,loaderName,"getClassLoader"));
		return mbs.getClassLoader(loaderName);
	}

	public ClassLoader getClassLoaderFor(ObjectName mbeanName)
			throws InstanceNotFoundException {
			accessController.checkPermission(new MBeanPermission(getClassName(mbeanName),null,mbeanName,"getClassLoaderFor"));
		return mbs.getClassLoaderFor(mbeanName);
	}

	public ClassLoaderRepository getClassLoaderRepository() {
			accessController.checkPermission(new MBeanPermission(null,null,null,"getClassLoaderRepository"));
		return mbs.getClassLoaderRepository();
	}

	public String getDefaultDomain() {
		return mbs.getDefaultDomain();
	}

	public String[] getDomains() {
		MBeanPermission perm = null;
		try {
			perm = new MBeanPermission(null,null,new ObjectName("*"),"getDomains");
		} catch (MalformedObjectNameException e) {
			logger.severe(e.getMessage());
		} catch (NullPointerException e) {
			logger.severe(e.getMessage());
		}
		accessController.checkPermission(perm);
		List<String> domainsList = Arrays.asList(mbs.getDomains());
		for(int i = 0;i<domainsList.size();i++){
			String domain = domainsList.get(i);
			try{
				accessController.checkPermission(new MBeanPermission(null,null, new ObjectName(domain+":x=x"), "getDomains"));
			}catch(AccessControlException ace){
				domainsList.remove(domain);
				i--;
			} catch (MalformedObjectNameException e) {
				logger.severe(e.getMessage());
			} catch (NullPointerException e) {
				logger.severe(e.getMessage());
			}
		}
		return domainsList.toArray(new String[domainsList.size()]);
	}

	public Integer getMBeanCount() {
		return mbs.getMBeanCount();
	}

	public MBeanInfo getMBeanInfo(ObjectName name)
			throws InstanceNotFoundException, IntrospectionException,
			ReflectionException {
		accessController.checkPermission(new MBeanPermission(getClassName(name),null,name,"getMBeanInfo"));
		return mbs.getMBeanInfo(name);
	}

	public ObjectInstance getObjectInstance(ObjectName name)
			throws InstanceNotFoundException {
		accessController.checkPermission(new MBeanPermission(getClassName(name),null,name,"getObjectInstance"));
		return mbs.getObjectInstance(name);
	}

	public Object instantiate(String className) throws ReflectionException,
			MBeanException {
		accessController.checkPermission(new MBeanPermission(className,null,null,"instantiate"));
		return mbs.instantiate(className);
	}

	public Object instantiate(String className, ObjectName loaderName)
			throws ReflectionException, MBeanException,
			InstanceNotFoundException {
		accessController.checkPermission(new MBeanPermission(className,null,null,"instantiate"));
		return mbs.instantiate(className,loaderName);
	}

	public Object instantiate(String className, Object[] params,
			String[] signature) throws ReflectionException, MBeanException {
		accessController.checkPermission(new MBeanPermission(className,null,null,"instantiate"));
		return mbs.instantiate(className, params, signature);
	}

	public Object instantiate(String className, ObjectName loaderName,
			Object[] params, String[] signature) throws ReflectionException,
			MBeanException, InstanceNotFoundException {
		accessController.checkPermission(new MBeanPermission(className,null,null,"instantiate"));
		return mbs.instantiate(className, loaderName, params, signature);
	}

	public Object invoke(ObjectName name, String operationName,
			Object[] params, String[] signature)
			throws InstanceNotFoundException, MBeanException,
			ReflectionException {
		accessController.checkPermission(new MBeanPermission(getClassName(name),operationName,name,"invoke"));
		return mbs.invoke(name, operationName, params, signature);
	}

	public boolean isInstanceOf(ObjectName name, String className)
			throws InstanceNotFoundException {
		accessController.checkPermission(new MBeanPermission(className,null,name,"isInstanceOf"));
		return mbs.isInstanceOf(name, className);
	}

	public boolean isRegistered(ObjectName name) {
		return mbs.isRegistered(name);
	}

	public Set queryMBeans(ObjectName name, QueryExp query) {
		accessController.checkPermission(new MBeanPermission(null,null,name,"queryMBeans"));
		Set<ObjectInstance> mbeans = mbs.queryMBeans(name, query);
		
		Set<ObjectInstance> mbeansToRemove = new HashSet<ObjectInstance>(); 
		for(ObjectInstance oi:mbeans){
			try{
				accessController.checkPermission(new MBeanPermission(oi.getClassName(),null,oi.getObjectName(),"queryMBeans"));
			}catch(AccessControlException ace){
				mbeansToRemove.add(oi);
			}
		}
		boolean success = mbeans.removeAll(mbeansToRemove);
		if(!success){
			throw new AccessControlException("mbeans cannot be removed from the returned Set when access is denied to them with the queryMbeans operation ");
		}
		return mbeans;
	}

	public Set queryNames(ObjectName name, QueryExp query) {
		accessController.checkPermission(new MBeanPermission(null,null,name,"queryNames"));
		Set<ObjectName> mbeans = mbs.queryNames(name, query);
		Set<ObjectName> mbeansToRemove = new HashSet<ObjectName>(); 
		for(ObjectName on:mbeans){
			try{
				accessController.checkPermission(new MBeanPermission(getClassName(on),null,on,"queryNames"));
			}catch(AccessControlException ace){
				mbeansToRemove.add(on);
			}
		}
		boolean success = mbeans.removeAll(mbeansToRemove);
		
		return mbeans;
	}

	public ObjectInstance registerMBean(Object object, ObjectName name)
			throws InstanceAlreadyExistsException, MBeanRegistrationException,
			NotCompliantMBeanException {
		String className = null;
		try {
			className = mbs.getMBeanInfo(name).getClassName();
		} catch (InstanceNotFoundException e) {
			logger.severe(e.getMessage());
		} catch (IntrospectionException e) {
			logger.severe(e.getMessage());
		} catch (ReflectionException e) {
			logger.severe(e.getMessage());
		}
		//instantiate permission check
		accessController.checkPermission(new MBeanPermission(className,null,null,"instantiate"));
		
		//register permission check
		if(name!=null){
			accessController.checkPermission(new MBeanPermission(className,null,name,"registerMBean"));
		}else{
			MBeanRegistration mbeanReg = (MBeanRegistration)object;
			try {
				name = (ObjectName)mbeanReg.preRegister(mbs, null);
			} catch (Exception e) {
				logger.severe(e.getMessage());
			}
			accessController.checkPermission(new MBeanPermission(className,null,name,"registerMBean"));
		}
		Class clazz = null;
		try {
			clazz = Thread.currentThread().getContextClassLoader().loadClass(className);
		} catch (ClassNotFoundException e) {
			logger.severe(e.getMessage());
		}
		 if(!clazz.getProtectionDomain().implies(new MBeanTrustPermission("register"))){
			 throw new AccessControlException("registration denied");
		 }
		return mbs.registerMBean(object, name);
	}

	public void removeNotificationListener(ObjectName name, ObjectName listener)
			throws InstanceNotFoundException, ListenerNotFoundException {
		accessController.checkPermission(new MBeanPermission(getClassName(name),null,name,"removeNotificationListener"));
		mbs.removeNotificationListener(name, listener);

	}

	public void removeNotificationListener(ObjectName name,
			NotificationListener listener) throws InstanceNotFoundException,
			ListenerNotFoundException {
		accessController.checkPermission(new MBeanPermission(getClassName(name),null,name,"removeNotificationListener"));
		mbs.removeNotificationListener(name, listener);

	}

	public void removeNotificationListener(ObjectName name,
			ObjectName listener, NotificationFilter filter, Object handback)
			throws InstanceNotFoundException, ListenerNotFoundException {
		accessController.checkPermission(new MBeanPermission(getClassName(name),null,name,"removeNotificationListener"));
		mbs.removeNotificationListener(name, listener,filter,handback);

	}

	public void removeNotificationListener(ObjectName name,
			NotificationListener listener, NotificationFilter filter,
			Object handback) throws InstanceNotFoundException,
			ListenerNotFoundException {
		accessController.checkPermission(new MBeanPermission(getClassName(name),null,name,"removeNotificationListener"));
		mbs.removeNotificationListener(name, listener,filter,handback);

	}

	public void setAttribute(ObjectName name, Attribute attribute)
			throws InstanceNotFoundException, AttributeNotFoundException,
			InvalidAttributeValueException, MBeanException, ReflectionException {
		accessController.checkPermission(new MBeanPermission(getClassName(name),attribute.getName(),name,"setAttribute"));

	}

	public AttributeList setAttributes(ObjectName name, AttributeList attributes)
			throws InstanceNotFoundException, ReflectionException {
		accessController.checkPermission(new MBeanPermission(getClassName(name),null,name,"setAttribute"));
		AttributeList attList = mbs.setAttributes(name, attributes);
		for(int i = 0;i<attList.size();i++){
			Attribute att = (Attribute)attList.get(i);
			try{
				accessController.checkPermission(new MBeanPermission(getClassName(name), att.getName(), name, "setAttribute"));
			}catch(AccessControlException ace){
				attList.remove(att);
				i--;
			}
		}
		return null;
	}

	public void unregisterMBean(ObjectName name)
			throws InstanceNotFoundException, MBeanRegistrationException {
		accessController.checkPermission(new MBeanPermission(getClassName(name),null,name,"unregisterMBean"));
		mbs.unregisterMBean(name);

	}

	private String getClassName(ObjectName name){
		String className = null;
		try {
			className = mbs.getMBeanInfo(name).getClassName();
		} catch (InstanceNotFoundException e) {
			logger.severe(e.getMessage());
		} catch (IntrospectionException e) {
			logger.severe(e.getMessage());
		} catch (ReflectionException e) {
			logger.severe(e.getMessage());
		}
		return className;
	}
}
